import os
import pathlib
import shutil
import tempfile
from typing import Tuple
import zipfile

from isic_challenge_scoring.types import ScoreException


def extract_zip(zip_path: pathlib.Path, output_path: pathlib.Path, flatten: bool = True) -> None:
    """Extract a zip file, optionally flattening it into a single directory."""
    try:
        with zipfile.ZipFile(zip_path) as zf:
            if flatten:
                for member_info in zf.infolist():
                    member_name = member_info.filename
                    if member_name.startswith('__MACOSX'):
                        # Ignore Mac OS X metadata
                        continue

                    member_base_name = os.path.basename(member_name)
                    if not member_base_name:
                        # Skip directories
                        continue

                    member_output_path = output_path / member_base_name

                    with zf.open(member_info) as input_stream, member_output_path.open(
                        'wb'
                    ) as output_stream:
                        shutil.copyfileobj(input_stream, output_stream)
            else:
                zf.extractall(output_path)
    except zipfile.BadZipfile as e:
        raise ScoreException(f'Could not read ZIP file "{zip_path.name}": {str(e)}.')


def unzip_all(input_file: pathlib.Path) -> Tuple[pathlib.Path, tempfile.TemporaryDirectory]:
    """Extract a ZIP file to a temporary directory."""
    output_temp_dir = tempfile.TemporaryDirectory()
    output_path = pathlib.Path(output_temp_dir.name)

    if input_file.suffix.lower() != '.zip':
        raise Exception(f'ZIP file expected at: "{str(input_file)}".')

    extract_zip(input_file, output_path)

    return output_path, output_temp_dir



import warnings

import numpy as np
import pandas as pd
import sklearn.metrics


def _to_labels(probabilities: pd.DataFrame) -> pd.Series:
    labels = probabilities.idxmax(axis='columns')

    # Find places where there are multiple maximum values
    max_probabilities = probabilities.max(axis='columns')
    is_max: pd.DataFrame = probabilities.eq(max_probabilities, axis='rows')
    number_of_max: pd.Series = is_max.sum(axis='columns')
    multiple_max: pd.Series = number_of_max.gt(1)
    # Set those locations as an 'undecided' label
    labels[multiple_max] = 'undecided'
    # TODO: emit a warning if any are set to 'undecided'

    return labels


def _get_frequencies(labels: pd.Series, weights: pd.Series, categories: pd.Index) -> pd.Series:
    # Directly sum the weights, grouping them by label
    # .reindex sorts this by the order in categories
    return weights.groupby(labels, sort=False).sum().reindex(categories, fill_value=0)


def _label_balanced_multiclass_accuracy(
    truth_labels: pd.Series, prediction_labels: pd.Series, weights: pd.Series, categories: pd.Index
) -> float:
    # See http://scikit-learn.org/dev/modules/model_evaluation.html#balanced-accuracy-score ; in
    # summary, 'sklearn.metrics.balanced_accuracy_score' is for binary classification only, so we
    # need to implement our own; here, we implement a simpler version of "balanced accuracy" than
    # the definitions mentioned by SciKit learn, as it's just a normalization of TP scores by true
    # class proportions
    confusion_matrix = sklearn.metrics.confusion_matrix(
        truth_labels, prediction_labels, labels=categories, sample_weight=weights
    )
    # TODO: try to convert to a DataFrame, for useful debugging labels
    # confusion_matrix = pd.DataFrame(
    #     confusion_matrix,
    #     index=LABELS.map(lambda label: f'true_{label}'),
    #     columns=LABELS.map(lambda label: f'predicted_{label}')
    # )

    tp_counts = pd.Series(confusion_matrix.diagonal(), index=categories)

    # These are equal to rows of the confusion matrix
    true_label_frequencies = _get_frequencies(truth_labels, weights, categories)

    balanced_accuracy = tp_counts.divide(true_label_frequencies).mean()
    return balanced_accuracy


def balanced_multiclass_accuracy(
    truth_probabilities: pd.DataFrame, prediction_probabilities: pd.DataFrame, weights: pd.Series
) -> float:
    truth_labels = _to_labels(truth_probabilities)
    prediction_labels = _to_labels(prediction_probabilities)
    categories = truth_probabilities.columns

    # This is easier to test
    balanced_accuracy = _label_balanced_multiclass_accuracy(
        truth_labels, prediction_labels, weights, categories
    )
    return balanced_accuracy


def binary_accuracy(cm: pd.Series) -> float:
    return (cm.at['TP'] + cm.at['TN']) / (cm.at['TP'] + cm.at['TN'] + cm.at['FP'] + cm.at['FN'])


def binary_sensitivity(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FN'] == 0:
        # sensitivity can't be calculated if all are negative, so make this metric a freebie
        return 1.0
    else:
        return cm.at['TP'] / (cm.at['TP'] + cm.at['FN'])


def binary_specificity(cm: pd.Series) -> float:
    if cm.at['TN'] + cm.at['FP'] == 0:
        # specificity can't be calculated if all are positive, so make this metric a freebie
        return 1.0
    else:
        return cm.at['TN'] / (cm.at['TN'] + cm.at['FP'])


def binary_jaccard(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FP'] + cm.at['FN'] == 0:
        # Jaccard is ill-defined if all are negative and the prediction is perfect, but we'll
        # just score that as a perfect answer
        return 1.0
    else:
        return cm.at['TP'] / (cm.at['TP'] + cm.at['FP'] + cm.at['FN'])


def binary_threshold_jaccard(cm: pd.Series, threshold: float = 0.65) -> float:
    jaccard = binary_jaccard(cm)
    return jaccard if jaccard >= threshold else 0.0


def binary_dice(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FP'] + cm.at['FN'] == 0:
        # Dice is ill-defined if all are negative and the prediction is perfect, but we'll
        # just score that as a perfect answer
        return 1.0
    else:
        return (2 * cm.at['TP']) / ((2 * cm.at['TP']) + cm.at['FP'] + cm.at['FN'])


def binary_ppv(cm: pd.Series) -> float:
    if cm.at['TP'] + cm.at['FP'] == 0:
        # PPV is ill-defined if all predictions are negative; we'll score it as perfect, which
        # doesn't penalize the case where all are truly negative (a good predictor), and is sane
        # for the case where some are truly positive (a limitation of this metric)
        # Note, some other implementations would score the latter case as 0:
        # https://github.com/dice-group/gerbil/wiki/Precision,-Recall-and-F1-measure
        return 1.0
    else:
        return cm.at['TP'] / (cm.at['TP'] + cm.at['FP'])


def binary_npv(cm: pd.Series) -> float:
    if cm.at['TN'] + cm.at['FN'] == 0:
        # NPV is ill-defined if all predictions are positive; we'll score it as perfect, which
        # doesn't penalize the case where all are truly positive (a good predictor), and is sane
        # for the case where some are truly negative (a limitation of this metric)
        # Note, some other implementations would score the latter case as 0:
        # https://github.com/dice-group/gerbil/wiki/Precision,-Recall-and-F1-measure
        return 1.0
    else:
        return cm.at['TN'] / (cm.at['TN'] + cm.at['FN'])


def auc(
    truth_probabilities: pd.Series, prediction_probabilities: pd.Series, weights: pd.Series
) -> float:
    auc = sklearn.metrics.roc_auc_score(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )
    return auc


def auc_above_sensitivity(
    truth_probabilities: pd.Series,
    prediction_probabilities: pd.Series,
    weights: pd.Series,
    sensitivity_threshold: float,
) -> float:
    if not (0 < sensitivity_threshold <= 1.0):
        raise Exception(f'Out of bounds sensitivity_threshold: {sensitivity_threshold}.')

    # Get the ROC curve points
    # TODO: We must have both some true and false instances in truthProbabilities
    fp_rates, tp_rates, thresholds = sklearn.metrics.roc_curve(
        truth_probabilities,
        prediction_probabilities,
        sample_weight=weights,
        drop_intermediate=False,
    )

    # Calling sklearn.metrics.roc_auc_score with max_fpr always applies the McClish correction,
    # which is a transform to normalize partial AUC values into the range [0.5, 1] (for a given FPR
    # interval): http://www.ncbi.nlm.nih.gov/pubmed/2668680
    # McClish-normalized partial AUC values may be a helpful metric to evaluate on their own,
    # but they are incompatible with the overall AUC, and SciKit learn (unlike R) does not
    # provide a flag to return the raw partial AUC, so just compute the desired metric directly

    # Search for the index along the curve where sensitivity_threshold (i.e. tp_rate threshold)
    # occurs
    # Since tp_rates is ordered, searchsorted provides better performance than np.argmax
    # Use side='left' to include any following points with exactly the target value
    threshold_index = tp_rates.searchsorted(sensitivity_threshold, side='left')

    # Take only the segment >= the value at threshold_index
    tp_rates_segment = tp_rates[threshold_index:]
    fp_rates_segment = fp_rates[threshold_index:]

    # Create an additional ROC point at exactly the threshold value
    tp_rate_threshold = sensitivity_threshold
    # It will be the case that fp_rate_threshold <= tp_rates[threshold_index]
    # Since tp_rates may have repeated values (which is disallowed by np.interp), use a 2-value
    # segment directly around the threshold value
    # If fp_rate_threshold < tp_rates[threshold_index], the 2-value segment needs to start from the
    # location at threshold_index-1, so that it straddles fp_rate_threshold
    # Even if fp_rate_threshold == tp_rates[threshold_index], a 2-value segment starting before
    # threshold_index is guaranteed to have no duplicates, as threshold_index is the left
    # side of any series of duplicates (since it was found with searchsorted(..., side='left'))
    fp_rate_threshold = np.interp(
        tp_rate_threshold,
        tp_rates[threshold_index - 1 : threshold_index + 1],
        fp_rates[threshold_index - 1 : threshold_index + 1],
    )

    # Prepend the point to the segment
    tp_rates_segment = np.insert(tp_rates_segment, 0, tp_rate_threshold)
    fp_rates_segment = np.insert(fp_rates_segment, 0, fp_rate_threshold)

    partial_auc = sklearn.metrics.auc(fp_rates_segment, tp_rates_segment)
    return partial_auc


def average_precision(
    truth_probabilities: pd.Series, prediction_probabilities: pd.Series, weights: pd.Series
) -> float:
    with warnings.catch_warnings():
        # sklearn.metrics.average_precision_score sometimes causes warnings internally, but they
        # appear to be harmless
        warnings.filterwarnings(
            'ignore', category=RuntimeWarning, message=r'^invalid value encountered in true_divide$'
        )
        ap = sklearn.metrics.average_precision_score(
            truth_probabilities, prediction_probabilities, sample_weight=weights
        )
    return ap


def roc(
    truth_probabilities: pd.Series, prediction_probabilities: pd.Series, weights: pd.Series
) -> pd.DataFrame:
    fprs, tprs, thresholds = sklearn.metrics.roc_curve(
        truth_probabilities, prediction_probabilities, sample_weight=weights
    )

    roc = pd.DataFrame({'fpr': fprs, 'tpr': tprs}, index=thresholds, columns=['fpr', 'tpr'])
    return roc



from isic_challenge_scoring.classification import ClassificationScore
from isic_challenge_scoring.segmentation import SegmentationScore
from isic_challenge_scoring.types import ScoreException

__all__ = ['ClassificationScore', 'SegmentationScore', 'ScoreException']



from __future__ import annotations

from dataclasses import dataclass
import pathlib
from typing import cast, Iterable

import pandas as pd

from isic_challenge_scoring import metrics
from isic_challenge_scoring.confusion import create_binary_confusion_matrix
from isic_challenge_scoring.load_image import ImagePair, iter_image_pairs
from isic_challenge_scoring.types import Score, ScoreDict, SeriesDict
from isic_challenge_scoring.unzip import unzip_all


@dataclass(init=False)
class SegmentationScore(Score):
    macro_average: pd.Series

    def __init__(self, image_pairs: Iterable[ImagePair]) -> None:
        # TODO: Add weighting
        confusion_matrics = pd.DataFrame(
            [
                create_binary_confusion_matrix(
                    truth_binary_values=image_pair.truth_image > 128,
                    prediction_binary_values=image_pair.prediction_image > 128,
                    name=image_pair.image_id,
                )
                for image_pair in image_pairs
            ]
        )

        per_image = pd.DataFrame(
            {
                'accuracy': confusion_matrics.apply(metrics.binary_accuracy, axis='columns'),
                'sensitivity': confusion_matrics.apply(metrics.binary_sensitivity, axis='columns'),
                'specificity': confusion_matrics.apply(metrics.binary_specificity, axis='columns'),
                'jaccard': confusion_matrics.apply(metrics.binary_jaccard, axis='columns'),
                'threshold_jaccard': confusion_matrics.apply(
                    metrics.binary_threshold_jaccard, threshold=0.65, axis='columns'
                ),
                'dice': confusion_matrics.apply(metrics.binary_dice, axis='columns'),
            },
            columns=[
                'accuracy',
                'sensitivity',
                'specificity',
                'jaccard',
                'threshold_jaccard',
                'dice',
            ],
        )

        self.macro_average = per_image.mean(axis='index').rename('macro_average', inplace=True)

        self.overall = self.macro_average.at['threshold_jaccard']
        self.validation = self.macro_average.at['threshold_jaccard']

    def to_string(self) -> str:
        output = super().to_string()
        output += '\n\nMacro averaged metrics:\n'
        output += self.macro_average.to_string()
        return output

    def to_dict(self) -> ScoreDict:
        output = super().to_dict()
        output.update({'macro_average': cast(SeriesDict, self.macro_average.to_dict())})
        return output

    @classmethod
    def from_dir(cls, truth_path: pathlib.Path, prediction_path: pathlib.Path) -> SegmentationScore:
        image_pairs = iter_image_pairs(truth_path, prediction_path)
        return cls(image_pairs)

    @classmethod
    def from_zip_file(
        cls, truth_zip_file: pathlib.Path, prediction_zip_file: pathlib.Path
    ) -> SegmentationScore:
        truth_path, truth_temp_dir = unzip_all(truth_zip_file)
        # TODO: If an exception occurs while unzipping prediction_zip_file, truth_temp_dir is not
        #  cleaned up
        prediction_path, prediction_temp_dir = unzip_all(prediction_zip_file)

        try:
            score = cls.from_dir(truth_path, prediction_path)
        finally:
            truth_temp_dir.cleanup()
            prediction_temp_dir.cleanup()

        return score



import pathlib
from typing import cast, Dict

import pandas as pd

from isic_challenge_scoring import metrics
from isic_challenge_scoring.confusion import (
    create_binary_confusion_matrix,
    normalize_confusion_matrix,
)
from isic_challenge_scoring.load_image import iter_image_pairs


def score(truth_path: pathlib.Path, prediction_path: pathlib.Path):
    confusion_matrics = pd.DataFrame(
        [
            create_binary_confusion_matrix(
                truth_binary_values=image_pair.truth_image > 128,
                prediction_binary_values=image_pair.prediction_image > 128,
                name=(cast(str, image_pair.attribute_id), image_pair.image_id),
            )
            for image_pair in iter_image_pairs(truth_path, prediction_path)
        ]
    )
    confusion_matrics = confusion_matrics.reindex(
        index=pd.MultiIndex.from_tuples(confusion_matrics.index, names=('attribute_id', 'image_id'))
    )

    # Normalize all values, since image sizes vary
    normalized_confusion_matrics = confusion_matrics.apply(
        normalize_confusion_matrix, axis='columns'
    )

    scores: Dict = {}
    for attribute in sorted(confusion_matrics.index.unique('attribute_id')):
        attribute_confusion_matrics = normalized_confusion_matrics.loc(axis=0)[attribute, :]
        sum_attribute_confusion_matrics = attribute_confusion_matrics.sum(axis='index')

        scores[attribute] = {
            'jaccard': metrics.binary_jaccard(sum_attribute_confusion_matrics),
            'dice': metrics.binary_dice(sum_attribute_confusion_matrics),
        }

    sum_confusion_matrix = normalized_confusion_matrics.sum(axis='index')
    scores['micro_average'] = {
        'jaccard': metrics.binary_jaccard(sum_confusion_matrix),
        'dice': metrics.binary_dice(sum_confusion_matrix),
    }

    scores['overall'] = scores['micro_average']['jaccard']

    return scores
